/**
 * eobjects.org MetaModel
 * Copyright (C) 2010 eobjects.org
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution; if not, write to:
 * Free Software Foundation, Inc.
 * 51 Franklin Street, Fifth Floor
 * Boston, MA  02110-1301  USA
 */
package org.eobjects.metamodel.data;

import java.util.Arrays;
import java.util.List;

import org.eobjects.metamodel.query.SelectItem;
import org.eobjects.metamodel.schema.Column;
import org.eobjects.metamodel.util.BaseObject;

/**
 * Default Row implementation. Holds values in memory.
 * 
 * @author Kasper Sørensen
 */
public final class DefaultRow extends BaseObject implements Cloneable, Row {

    private static final long serialVersionUID = 1L;

    private final DataSetHeader _header;
    private final Object[] _values;
    private final Style[] _styles;

    /**
     * This field was replaced by the DataSetHeader field above.
     * 
     * @deprecated no longer used, except for in deserialized objects
     */
    @Deprecated
    private List<SelectItem> _items;

    /**
     * Constructs a row.
     * 
     * @param header
     * @param values
     * @param styles
     */
    public DefaultRow(DataSetHeader header, Object[] values, Style[] styles) {
        if (header == null) {
            throw new IllegalArgumentException("DataSet header cannot be null");
        }
        if (values == null) {
            throw new IllegalArgumentException("Values cannot be null");
        }
        if (header.size() != values.length) {
            throw new IllegalArgumentException("Header size and values length must be equal. " + header.size()
                    + " select items present in header and encountered these values: " + Arrays.toString(values));
        }
        if (styles != null) {
            if (values.length != styles.length) {
                throw new IllegalArgumentException("Values length and styles length must be equal. " + values.length
                        + " values present and encountered these styles: " + Arrays.toString(styles));
            }
            boolean entirelyNoStyle = true;
            for (int i = 0; i < styles.length; i++) {
                if (styles[i] == null) {
                    throw new IllegalArgumentException("Elements in the style array cannot be null");
                }
                if (entirelyNoStyle && !Style.NO_STYLE.equals(styles[i])) {
                    entirelyNoStyle = false;
                }
            }

            if (entirelyNoStyle) {
                // no need to reference any styles
                styles = null;
            }
        }
        _header = header;
        _values = values;
        _styles = styles;
    }

    /**
     * Constructs a row.
     * 
     * @param header
     * @param values
     */
    public DefaultRow(DataSetHeader header, Object[] values) {
        this(header, values, null);
    }

    /**
     * Constructs a row from an array of SelectItems and an array of
     * corresponding values
     * 
     * @param items
     *            the array of SelectItems
     * @param values
     *            the array of values
     * 
     * @deprecated use {@link #DefaultRow(DataSetHeader, Object[])} or
     *             {@link #DefaultRow(DataSetHeader, Object[], Style[])}
     *             instead.
     */
    @Deprecated
    public DefaultRow(SelectItem[] items, Object[] values) {
        this(Arrays.asList(items), values, null);
    }

    /**
     * Constructs a row from an array of SelectItems and an array of
     * corresponding values
     * 
     * @param items
     *            the array of SelectItems
     * @param values
     *            the array of values
     * @param styles
     *            an optional array of styles
     * @deprecated use {@link #DefaultRow(DataSetHeader, Object[])} or
     *             {@link #DefaultRow(DataSetHeader, Object[], Style[])}
     *             instead.
     */
    @Deprecated
    public DefaultRow(SelectItem[] items, Object[] values, Style[] styles) {
        this(Arrays.asList(items), values, styles);
    }

    /**
     * Constructs a row from a list of SelectItems and an array of corresponding
     * values
     * 
     * @param items
     *            the list of SelectItems
     * @param values
     *            the array of values
     * @deprecated use {@link #DefaultRow(DataSetHeader, Object[])} or
     *             {@link #DefaultRow(DataSetHeader, Object[], Style[])}
     *             instead.
     */
    @Deprecated
    public DefaultRow(List<SelectItem> items, Object[] values) {
        this(items, values, null);
    }

    /**
     * Constructs a row from a list of SelectItems and an array of corresponding
     * values
     * 
     * @param items
     *            the list of SelectItems
     * @param values
     *            the array of values
     * @param styles
     *            an optional array of styles
     * @deprecated use {@link #DefaultRow(DataSetHeader, Object[])} or
     *             {@link #DefaultRow(DataSetHeader, Object[], Style[])}
     *             instead.
     */
    @Deprecated
    public DefaultRow(List<SelectItem> items, Object[] values, Style[] styles) {
        this(new SimpleDataSetHeader(items), values, styles);
    }

    @Override
    public Object getValue(SelectItem item) {
        if (item == null) {
            return null;
        }

        int index = getHeader().indexOf(item);
        if (index == -1) {
            return null;
        }
        return getValue(index);
    }

    @Override
    public Object getValue(Column column) {
        int index = indexOf(column);
        if (index == -1) {
            return null;
        }
        return getValue(index);
    }

    @Override
    public int indexOf(SelectItem item) {
        return getHeader().indexOf(item);
    }

    @Override
    public int indexOf(Column column) {
        return getHeader().indexOf(column);
    }

    @Override
    public Object getValue(int index) throws ArrayIndexOutOfBoundsException {
        return _values[index];
    }

    @Override
    public SelectItem[] getSelectItems() {
        return getHeader().getSelectItems();
    }

    @Override
    public Object[] getValues() {
        return _values;
    }

    @Override
    public String toString() {
        return "Row[values=" + Arrays.toString(_values) + "]";
    }

    @Override
    protected void decorateIdentity(List<Object> identifiers) {
        identifiers.add(getHeader());
        identifiers.add(_values);
        identifiers.add(_styles);
    }

    @Override
    protected DefaultRow clone() {
        return new DefaultRow(getHeader(), _values, _styles);
    }

    @Override
    public Row getSubSelection(DataSetHeader header) {
        final int size = header.size();
        final Object[] values = new Object[size];
        final Style[] styles = new Style[size];
        for (int i = 0; i < size; i++) {
            SelectItem selectItem = header.getSelectItem(i);

            if (selectItem.getSubQuerySelectItem() != null) {
                values[i] = getValue(selectItem.getSubQuerySelectItem());
                styles[i] = getStyle(selectItem.getSubQuerySelectItem());
                if (values[i] == null) {
                    values[i] = getValue(selectItem);
                    styles[i] = getStyle(selectItem);
                }
            } else {
                values[i] = getValue(selectItem);
                styles[i] = getStyle(selectItem);
            }
        }
        return new DefaultRow(header, values, styles);
    }

    @Override
    public Row getSubSelection(final SelectItem[] selectItems) {
        final DataSetHeader header = new SimpleDataSetHeader(selectItems);
        return getSubSelection(header);
    }

    @Override
    public Style getStyle(SelectItem item) {
        if (_styles == null) {
            return Style.NO_STYLE;
        }
        return _styles[indexOf(item)];
    }

    @Override
    public Style getStyle(Column column) {
        if (_styles == null) {
            return Style.NO_STYLE;
        }
        return _styles[indexOf(column)];
    }

    @Override
    public Style getStyle(int index) throws IndexOutOfBoundsException {
        if (_styles == null) {
            return Style.NO_STYLE;
        }
        return _styles[index];
    }

    @Override
    public int size() {
        return getHeader().size();
    }

    private DataSetHeader getHeader() {
        if (_header == null && _items != null) {
            // this only happens for deserialized objects which where serialized
            // prior to the introduction of DataSetHeader.
            return new SimpleDataSetHeader(_items);
        }
        return _header;
    }
}